import unittest
import time
import networkx as nx
from shapG.shapley import shapG, shapley_value, graph_generator

class TestShapleyValue(unittest.TestCase):
    shapley_values = {}
    shapley_values_G = {}
    @classmethod
    def setUp(cls):
        cls.G = graph_generator(10, 0.5)

    def test_shapley_value_degree_local(self):
        print("Shapley value start running...")
        start = time.time()
        TestShapleyValue.shapley_values = shapley_value(TestShapleyValue.G)
        end = time.time()
        print("Shapley value time: ", end - start)
        self.assertIsInstance(TestShapleyValue.shapley_values, dict)

    def test_shapG(self):
        print("ShapG start running...")
        start = time.time()
        TestShapleyValue.shapley_values_G = shapG(TestShapleyValue.G, approximate_by_ratio=False)
        end = time.time()
        print("ShapG running time: ", end - start)
        self.assertIsInstance(TestShapleyValue.shapley_values_G, dict)
    
    @classmethod
    def tearDown(cls):
        if cls.shapley_values and cls.shapley_values_G:
            print("\nComparing Shapley values...")
            print(f"{'Node':>5} | {'SV':>10} | {'ASV':>10} | {'diff':>10}")
            print("-" * 45)
            for node in cls.shapley_values:
                sv = cls.shapley_values[node]
                asv = cls.shapley_values_G.get(node, 'N/A')
                diff = 'N/A' if asv == 'N/A' else asv - sv
                print(f"{node:>5} | {sv:>10.4f} | {asv:>10.4f} | {diff:.2e}")



if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np
import networkx as nx
from scipy.stats import pearsonr,kendalltau,spearmanr
from sklearn.metrics import mutual_info_score

import logging
def corr_generator(df, method=kendalltau):
    """generate the correlation matrix of a dataframe

    Args:
        df (pd.DataFrame): DataFrame
        method (function, optional): the function to calculate the correlation coefficient. Defaults to pearsonr.
                                     it canbe pearsonr, kendalltau, spearmanr
    Returns:
        _type_: pd.DataFrame: correlation matrix
    """
    if method not in [pearsonr, kendalltau, spearmanr]:
        raise ValueError("method should be pearsonr, kendalltau, spearmanr")
    # Initialize dataframes for storing correlation values
    corr_df = pd.DataFrame(np.zeros((df.shape[1], df.shape[1])), columns=df.columns, index=df.columns)
    # Iterate over all pairs of columns
    for col1 in df.columns:
        for col2 in df.columns:
            if col1 != col2:  # Ensuring not to correlate a column with itself
                corr, _ = method(df[col1], df[col2])
                corr_df.loc[col1, col2] = corr
    return corr_df

def matrix_generator(df, method=kendalltau):
    """generate the correlation matrix of a dataframe

    Args:
        df (pd.DataFrame): DataFrame
        method (function, optional): the function to calculate the correlation coefficient. Defaults to pearsonr.
                                     it canbe pearsonr, kendalltau, spearmanr, mutual information (mutual_info_score), K-L divergence(kl)
    Returns:
        _type_: pd.DataFrame: correlation matrix
    """
    if method in [pearsonr, kendalltau, spearmanr]:
        return corr_generator(df, method)
    else:
        # Initialize dataframes for storing correlation values
        matrix_df = pd.DataFrame(np.zeros((df.shape[1], df.shape[1])), columns=df.columns, index=df.columns)
        # Iterate over all pairs of columns
        for col1 in matrix_df.columns:
            for col2 in matrix_df.columns:
                if col1 != col2:  # Ensuring not to correlate a column with itself
                    measures = method(df[col1], df[col2])
                    matrix_df.loc[col1, col2] = measures
    return matrix_df

def kl(P,Q):
    """ Epsilon is used here to avoid conditional code for
    checking that neither P nor Q is equal to 0. """
    epsilon = 1e-10
    P = P+epsilon
    Q = Q+epsilon
    P = P / np.sum(P)
    Q = Q / np.sum(Q)
    divergence = np.sum(P*np.log(P/Q))
    return divergence

def create_minimal_edge_graph(W, version='v3', reverse=True, verbose=False):
    """weight matrix to reduced adjacency matrix

    Args:
        W (pd.DataFrame): DataFrame
        version (str): version of the algorithm
            v1: stop when all nodes are added to the graph
            v2: when all nodes are added to the graph, then check the connectivity, 
                continue add edges until the graph is connected
            v3: strong connectivity check, more edges than v2
            example: 
            links: [1,2] [1,3] [2,3] [4,5] [4,6] [5,6] [1,4]
            v1: [1,2] [1,3] [4,5] [4,6]
            v2: [1,2] [1,3] [4,5] [4,6] [1,4]
            v3: [1,2] [1,3] [2,3] [4,5] [4,6] [5,6] [1,4]
        reverse (bool): sorting order (True: descending, False: ascending)

    Returns:
        adjacency_matrix: reduced adjacency matrix
        reduced_df: reduced weight matrix
    """
    edges = []
    columns = W.columns.tolist()
    for i in range(len(columns)):
        for j in range(i + 1, len(columns)):
            edges.append((columns[i], columns[j], abs(W.iloc[i, j])))

    # sorting
    edges.sort(key=lambda x: x[2], reverse=reverse)

    connected_nodes = set()
    adjacency_matrix = pd.DataFrame(0, index=columns, columns=columns, dtype=np.int8)
    reduced_df = pd.DataFrame(0, index=columns, columns=columns, dtype=np.float64)
    # add edges to the graph until all nodes are connected
    for edge in edges:
        node1, node2, weight = edge
        if node1 not in connected_nodes or node2 not in connected_nodes:
            adjacency_matrix.loc[node1, node2] = 1
            adjacency_matrix.loc[node2, node1] = 1
            reduced_df.loc[node1, node2] = weight
            reduced_df.loc[node2, node1] = weight
            connected_nodes.update([node1, node2])
        if version == 'v1':
            if len(connected_nodes) == len(columns):
                if verbose:
                    print(weight)
                break
        elif version == 'v2':
            if len(connected_nodes) == len(columns):
                G = nx.Graph(adjacency_matrix)
                if nx.is_connected(G):
                    if verbose:
                        print(weight)
                    break
                else:
                    adjacency_matrix.loc[node1, node2] = 1
                    adjacency_matrix.loc[node2, node1] = 1
                    reduced_df.loc[node1, node2] = weight
                    reduced_df.loc[node2, node1] = weight
        elif version == 'v3':
            G = nx.Graph(adjacency_matrix)
            if nx.is_connected(G) and len(connected_nodes) == len(columns):
                if verbose:
                    print(weight)
                break
            else:
                adjacency_matrix.loc[node1, node2] = 1
                adjacency_matrix.loc[node2, node1] = 1
                reduced_df.loc[node1, node2] = weight
                reduced_df.loc[node2, node1] = weight
                connected_nodes.update([node1, node2])

    return adjacency_matrix, reduced_df


import matplotlib.pyplot as plt
import numpy as np

def plot(shapley_values, top_n=10, style='seaborn-v0_8' ,file_name=None):
    """plot the shapley values

    Args:
        shapley_values (dict): dictionary of shapley values
        top_n (int, optional): if dictionary is too long, only plot top-n. Defaults to 10.
    """
    # sort in DESC order
    sorted_values = sorted(shapley_values.items(), key=lambda item: item[1], reverse=True)
    
    # select top-n
    if len(sorted_values) > top_n:
        sorted_values = sorted_values[:top_n]
    
    nodes, values = zip(*sorted_values)
    
    # set gnuplot style
    plt.style.use(style)
    
    # plot the bar chart
    fig, ax = plt.subplots()
    y_pos = np.arange(len(nodes))
    
    ax.barh(y_pos, values, align='center')
    ax.set_yticks(y_pos)
    ax.set_yticklabels(nodes)
    ax.invert_yaxis()  # 从上到下排列
    ax.set_xlabel('Shapley Value')
    ax.set_title('Top Shapley Values')
    
    # add the label
    for i, v in enumerate(values):
        ax.text(v + 0.01, i, f'{v:.2f}', va='center')
    if file_name is not None:
        plt.savefig(file_name, format='eps', dpi=300)
    plt.show()
    
if __name__ == '__main__':
    shapley_values = {
        0: 0.1,
        1: 0.2,
        2: 0.3,
        3: 0.4,
        4: 0.5,
        5: 0.6,
        6: 0.7,
        7: 0.8,
        8: 0.9,
        9: 1.0,
    }
    plot(shapley_values)


from setuptools import setup, find_packages

setup(
    name="shapG",
    version="0.13",
    packages=find_packages(),
    install_requires=[
        "networkx",
        "matplotlib",
        "numpy",
        "tqdm",
        "pandas",
        "scipy",
        "numpy",
    ],
    extras_require={
        "dev": [
            "unittest",
        ],
    },
    author="Chi Zhao",
    author_email="dandanv5@hotmail.com",
    description="A library to compute Shapley values in graphs.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/vectorsss/shapG",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import networkx as nx
import itertools
import random
from math import factorial, log2, ceil
from tqdm import tqdm

def graph_generator(n_nodes,density, weight_range=(1, 10),seed=2333):
    """generate a random graph based on the density

    Args:
        n_nodes (int): number of nodes
        density (float): the density of the graph
        weight_range (tuple, optional): range of weight. Defaults to (1, 10).
        seed (int, optional): random seed. Defaults to 2333.

    Returns:
        nx.Graph: graph
    """
    random.seed(seed)
    G = nx.Graph()
    G.add_nodes_from(range(n_nodes))  
    max_edges = n_nodes * (n_nodes - 1) / 2
    # calculate the expected number of edges
    n_edges = int(max_edges * density)

    while G.number_of_edges() < n_edges:
        u, v = random.sample(list(G.nodes()), 2)
        if not G.has_edge(u, v):
            if weight_range is None:
                G.add_edge(u, v)
            else:
                weight = random.randint(*weight_range)
                G.add_edge(u, v, weight=weight)

    return G

def coalition_degree(G, S):
    """characteristic function of a coalition in a graph

    Args:
        G (nx.Graph): the graph
        S (list/set): list/set of nodes to generate teh coalition

    Returns:
        int: the marginal contribution of the coalition
    """
    subgraph = G.subgraph(S)
    return sum(dict(subgraph.degree(weight='weight')).values()) / 2

def shapley_value(G: nx.Graph,f=coalition_degree):
    """shapley value without sampling

    Args:
        G (nx.Graph): graph
        f (function, optional): the characteristic function based on graph. Defaults to coalition_degree.

    Returns:
        dict: dictionary of shapley value
    """
    n_nodes = len(G.nodes())
    shapley_values = {node: 0 for node in G.nodes()}
    
    # Precompute factorials to improve efficiency
    fact = [factorial(i) for i in range(n_nodes + 1)]
    
    for node in tqdm(G.nodes()):
        for r in range(n_nodes + 1):
            for subset in itertools.combinations(G.nodes(), r):
                if node not in subset:
                    S = list(subset)
                    S_with_node = S + [node]
                    coeff = (fact[len(S)] * fact[n_nodes - len(S) - 1]) / fact[n_nodes]
                    marginal_contribution = f(G, S_with_node) - f(G, S)
                    shapley_values[node] += coeff * marginal_contribution
    return shapley_values

def get_neighbors_at_depth(G, node, depth):
    """Get k-hop neighbors for a given node.

    Args:
        G (nx.Graph): Graph.
        node (nx.Node): The given node.
        depth (int): Depth (k).

    Returns:
        list: k-hop neighbors.
    """
    path_lengths = nx.single_source_shortest_path_length(G, node, cutoff=depth)
    neighbors = {n for n, d in path_lengths.items() if d == depth}
    return neighbors

def shapG(G: nx.Graph, f=coalition_degree, depth=1, m=15, approximate_by_ratio=True, scale=True, verbose=False):
    """shapG algorithm with local search

    Args:
        G (nx.Graph): graph
        f (function, optional): characteristic function. Defaults to coalition_degree.
        depth (int, optional): the maximal depth. Defaults to 1.
        m (int, optional): the maximal neighbors. Defaults to 15.
        approximate_by_ratio (bool, optional): whether to approximate by ratio. Defaults to False (depends on the characteristic function).
        scale (bool, optional): the scaling factor. Defaults to True.

    Returns:
        dict: dictionary of shapley value
    """
    shapley_values = {node: 0 for node in G.nodes()}
    for node in tqdm(G.nodes()):
        neighbors_at_depth = set()
        for d in range(1, depth + 1):
            neighbors_at_depth |= get_neighbors_at_depth(G, node, d)
        
        if len(neighbors_at_depth) < m:
            neighbors_at_depth.add(node)
            for S_size in range(len(neighbors_at_depth) + 1):
                for S in itertools.combinations(neighbors_at_depth, S_size):
                    if node not in S:
                        S_with_node = list(S) + [node]
                        marginal_contribution = f(G, S_with_node) - f(G, S)
                        shapley_values[node] += marginal_contribution
            coeff = 1 / 2 ** (len(neighbors_at_depth) - 1)
            shapley_values[node] *= coeff
        else:
            # Eine Wahrscheinlichkeitsaufgabe in der Kundenwerbung Equation 18
            sample_nums = ceil(len(neighbors_at_depth) / m * (log2(len(neighbors_at_depth)) + 0.5772156649))  # Average sampling times
            for i in range(sample_nums):
                neighbors_at_depth_sampled = set(random.sample(list(neighbors_at_depth), m))
                neighbors_at_depth_sampled.add(node)
                iteration = 0
                for S_size in range(len(neighbors_at_depth_sampled) + 1):
                    for S in itertools.combinations(neighbors_at_depth_sampled, S_size):
                        if verbose and iteration % 100000 == 0:
                            print("{}: {}/{}".format(node, iteration, 1 << len(neighbors_at_depth_sampled)))
                        if node not in S:
                            S_with_node = list(S) + [node]
                            marginal_contribution = f(G, S_with_node) - f(G, S)
                            shapley_values[node] += marginal_contribution
                        iteration += 1
            coeff = 1 / 2 ** (len(neighbors_at_depth_sampled) - 1) / sample_nums
            if scale:
                coeff *= ((len(neighbors_at_depth) + 1) / len(neighbors_at_depth_sampled))
            shapley_values[node] *= coeff
    if approximate_by_ratio:
        biggest_coalition_contribution = f(G, G.nodes())
        approximate_biggest_coalition_contribution = sum(shapley_values.values())
        for k,v in shapley_values.items():
            corrected_v = v / approximate_biggest_coalition_contribution * biggest_coalition_contribution
            shapley_values[k] = corrected_v
    return shapley_values
